/*
Illusion of Gaia Map Viewer
Written by Alchemic
2010 May 28


This version only looks at the first layer for arrangements and 
tilesets. It works, but there are no backgrounds on the maps.

Incomplete
*/

using System;
using System.IO;
using System.Drawing;
using System.Windows.Forms;

public class IOGMapViewer : Form
{
	Button      btnLoadROM;
	Label       lblMapNumber;
	TextBox     txtMapNumber;
	Button      btnViewMap;
	Button      btnClear;
	PictureBox  pbxMap;
	Panel       pnlMap;

	int[]  powersOfTwo = { 0x80, 0x40, 0x20, 0x10, 0x8, 0x4, 0x2, 0x1 };

	byte[]    rom;
	byte[]    graphic;
	Color[]   palette;
	byte[][]  tileset;
	byte[][]  arrange;
	int[]     arrangeWidth;
	int[]     arrangeHeight;

	Bitmap[]  miniTiles;  // 8x8-pixel tiles
	Bitmap[]  megaTiles;  // 16x16-pixel tiles
	Bitmap    mapBitmap;





	void LoadROM(object sender, EventArgs e)
	{
		OpenFileDialog  romDialog;
		long            romLength;
		BinaryReader    romReader;


		romDialog = new OpenFileDialog();
		romDialog.Title = "Select an Illusion of Gaia ROM";
		romDialog.Filter = 
		  "SNES ROM files (*.smc)|*.smc" + 
		  "|All files (*.*)|*.*";

		if (romDialog.ShowDialog() == DialogResult.OK)
		{
			romLength = new FileInfo(romDialog.FileName).Length;
			if (((romLength & 0x7FFF) != 0x0) && ((romLength & 0x7FFF) != 0x200))
			{
				MessageBox.Show(
				  "The selected file had an unusual size and was not read.", 
				  "Unexpected File Size", 
				  MessageBoxButtons.OK, 
				  MessageBoxIcon.Exclamation
				);
			}
			else
			{
				rom = new byte[0x200000];
				romReader = new BinaryReader(File.Open(romDialog.FileName, FileMode.Open));
				romReader.BaseStream.Seek(romLength & 0x7FFF, SeekOrigin.Begin);
				romReader.Read(rom, 0, 0x200000);
				romReader.Close();
			}
		}
	}





	int FindMetadata(short goalMap, int index)
	{
		short  currentMap;


		while (true)
		{
			// Get the number of the current map.
			currentMap = BitConverter.ToInt16(rom, index);
			index += 2;

			// Is the current map the one we're looking for?
			if (currentMap == goalMap)
			{
				// If yes, return the current index.
				return index;
			}
			else
			{
				// If no, skip over the current map.
				while ((rom[index] != 0x00) && (rom[index] != 0x12))
				{
					switch (rom[index])
					{
						case 0x02:  index += 1;  break;
						case 0x03:  index += 7;  break; // Graphic
						case 0x04:  index += 6;  break; // Palette
						case 0x05:  index += 7;  break; // Tileset
						case 0x06:  index += 4;  break; // Arrangement
						case 0x10:  index += 6;  break;
						case 0x11:  index += 5;  break; // Music
						case 0x13:  index += 2;  break;
						case 0x14:  index += 1;  break;
						case 0x15:  index += 1;  break;
						case 0x17:  index += 4;  break;
					}
					index += 1;
				}

				// [00] indicates the end of the current map.
				// Continue with the next map.
				if (rom[index] == 0x00)
				{
					index += 1;
					continue;
				}

				// [12] indicates the end of the last map.
				// (This is a "map not found" case.)
				return -1;
			}
		}
	}





      void Decompress(int index, int length, out byte[] output)
	{
		byte[]  window;
		int     decompCount;


		// Create the output array.
		output = new byte[length];

		// The sliding window is initially filled with 0x20s.
		window = new byte[0x100];
		for (int i = 0; i < 0x100; i++)
		{
			window[i] = 0x20;
		}

		// Count bits instead of bytes.
		index *= 8;
		decompCount = 0;
		while (decompCount < length)
		{
			int nextAction = (rom[index/8] & powersOfTwo[index%8]) / powersOfTwo[index%8];
			index += 1;

			if (nextAction == 0)
			{
				// 0 bit: Pastcopy case.
				// The source-address counting starts with 0xEF. I have no idea why.
				int sourceCount = 0;
				for (int i = 0; i < 8; i++)
				{
					sourceCount *= 2;
					sourceCount += (rom[index/8] & powersOfTwo[index%8]) / powersOfTwo[index%8];
					index += 1;
				}

				int amountToCopy = 0;
				for (int i = 0; i < 4; i++)
				{
					amountToCopy *= 2;
					amountToCopy += (rom[index/8] & powersOfTwo[index%8]) / powersOfTwo[index%8];
					index += 1;
				}
				amountToCopy += 2;

				// Make sure the copy won't go off the end of the array.
				if ((decompCount + amountToCopy) >= length)
				{
					amountToCopy = length - decompCount;
				}

				// Copy the bytes.
				for (int i = 0; i < amountToCopy; i++)
				{
					output[decompCount] = window[sourceCount];
					window[(0xEF + decompCount) & 0xFF] = window[sourceCount];
					decompCount += 1;
					sourceCount = (sourceCount + 1) & 0xFF;
				}
			}
			else
			{
				// 1 bit: Literal case.
				int currentLiteral = 0;
				for (int i = 0; i < 8; i++)
				{
					currentLiteral *= 2;
					currentLiteral += (rom[index/8] & powersOfTwo[index%8]) / powersOfTwo[index%8];
					index += 1;
				}
				byte currentByte = (byte) (currentLiteral & 0xFF);

				output[decompCount] = currentByte;
				window[(0xEF + decompCount) & 0xFF] = currentByte;
				decompCount += 1;
			}
		}

		return;
	}





	int ReadPointer(int index)
	{
		int  pointer;


		pointer = BitConverter.ToInt32(rom, index);
		pointer &= 0x1FFFFF;

		return pointer;
	}





	void ReadGraphic(int index)
	{
		int  sourceStartIndex = (int) rom[index + 1];
		int  sourceEndIndex   = (int) rom[index + 2];
		int  destinationIndex = (int) rom[index + 3];
		int  sourcePointer    = ReadPointer(index + 4);
		int  unknownArgSeven  = (int) rom[index + 7];

		int     dataSize;
		byte[]  thisGraphic;


		switch (unknownArgSeven)
		{
			case 0x00:  destinationIndex += 0x20;  break;
			case 0x01:  destinationIndex += 0x40;  break;
			case 0x02:  destinationIndex += 0x60;  break;
		}

		dataSize = BitConverter.ToInt16(rom, sourcePointer);
		Decompress(sourcePointer + 2, dataSize, out thisGraphic);

		Buffer.BlockCopy(
		  thisGraphic,
		  0x200 * sourceStartIndex,
		  graphic,
		  0x200 * destinationIndex,
		  0x200 * (sourceEndIndex - sourceStartIndex)
		);

		return;
	}





	void ReadPalette(int index)
	{
		int  sourceStartIndex = (int) rom[index + 1];
		int  sourceEndIndex   = (int) rom[index + 2];
		int  destinationIndex = (int) rom[index + 3];
		int  sourcePointer    = ReadPointer(index + 4);

		short  currentColor;


		// Palettes are not compressed.
		for (int i = 0; i < sourceEndIndex - sourceStartIndex; i++)
		{
			currentColor = BitConverter.ToInt16(
			  rom, 
			  sourcePointer + (2 * (sourceStartIndex + i))
			);
			palette[destinationIndex + i] = Color.FromArgb(
			  ((currentColor & 0x001F) >>  0) * 8,
			  ((currentColor & 0x03E0) >>  5) * 8,
			  ((currentColor & 0x7C00) >> 10) * 8
			);
		}

		// The game does this... I have no idea why.
		if (destinationIndex == 0x10)
		{
			palette[0x00] = palette[0x10];
		}

		return;
	}





	void ReadTileset(int index)
	{
		int  sourceStartIndex = (int) rom[index + 1];
		int  sourceEndIndex   = (int) rom[index + 2];
		int  destinationIndex = (int) rom[index + 3];
		int  tilesetsToUpdate = (int) rom[index + 4];
		int  sourcePointer    = ReadPointer(index + 5);

		int     dataSize;
		byte[]  thisTileset;


		dataSize = BitConverter.ToInt16(rom, sourcePointer);
		Decompress(sourcePointer + 2, dataSize, out thisTileset);

		// Note that both tilesets can be updated at once.
		if ((tilesetsToUpdate & 0x01) != 0)
		{
			Buffer.BlockCopy(
			  thisTileset,
			  0x40 * sourceStartIndex,
			  tileset[0],
			  0x40 * destinationIndex,
			  0x40 * (sourceEndIndex - sourceStartIndex)
			);
		}

		if ((tilesetsToUpdate & 0x02) != 0)
		{
			Buffer.BlockCopy(
			  thisTileset,
			  0x40 * sourceStartIndex,
			  tileset[1],
			  0x40 * destinationIndex,
			  0x40 * (sourceEndIndex - sourceStartIndex)
			);
		}

		return;
	}





	void ReadArrange(int index)
	{
		int  layer    = (int) rom[index + 1];
		int  pointer  = ReadPointer(index + 2);
		int  width    = (int) rom[pointer + 0];
		int  height   = (int) rom[pointer + 1];
		int  dataSize = BitConverter.ToInt16(rom, pointer + 2);


		layer -= 1;
		arrangeWidth[layer]  = width;
		arrangeHeight[layer] = height;
		Decompress(pointer + 4, dataSize, out arrange[layer]);

		return;
	}





	void ViewMap(object sender, EventArgs e)
	{
		int    index;
		short  goalMap;


		if (rom == null)
		{
			MessageBox.Show(
			  "You have not loaded a ROM yet!", 
			  "No ROM Loaded", 
			  MessageBoxButtons.OK, 
			  MessageBoxIcon.Exclamation
			);

			return;
		}

		if (Int16.TryParse(txtMapNumber.Text, out goalMap) == false)
		{
			MessageBox.Show(
			  "The map number must be an integer.",
			  "Invalid Map Number",
			  MessageBoxButtons.OK, 
			  MessageBoxIcon.Exclamation
			);

			return;
		}

		index = FindMetadata(goalMap, 0xD8000);
		if (index == -1)
		{
			MessageBox.Show(
			  "Map " + goalMap + " could not be found.",
			  "Map Not Found",
			  MessageBoxButtons.OK, 
			  MessageBoxIcon.Exclamation
			);

			return;
		}


		// Parse the metadata.
		while ((rom[index] != 0x00) && (rom[index] != 0x12))
		{
			switch (rom[index])
			{
				case 0x02:
					index += 1;
					break;
				case 0x03:
					ReadGraphic(index);
					index += 7;
					break;
				case 0x04:
					ReadPalette(index);
					index += 6;
					break;
				case 0x05:
					ReadTileset(index);
					index += 7;
					break;
				case 0x06:
					ReadArrange(index);
					index += 4;
					break;
				case 0x10:
					index += 6;
					break;
				case 0x11:
					index += 5;
					break;
				case 0x13:
					index += 2;
					break;
				case 0x14:
					index += 1;
					break;
				case 0x15:
					index += 1;
					break;
				case 0x17:
					index += 4;
					break;
			}
			index += 1;
		}


		// Create the map image from the raw data.
		// This code only uses the first tileset and arrangement right now.

		mapBitmap = new Bitmap(
		  0x100 * arrangeWidth[0], 
		  0x100 * arrangeHeight[0]
		);


		// TILESET
		if (tileset[0] != null)
		{
			// MINI-TILES
			// Outer loop variables
			short  tileBytes;
			bool   tileVFlip;
			bool   tileHFlip;
			// bool   tilePriority;
			int    tilePalette;
			int    tileNumber;

			// Inner loop variables
			int    currentPixel;
			int    gIndex;
			int    pIndex;

			// Construct the minitiles
			for (int i = 0; i < 0x400; i++)
			{
				miniTiles[i] = new Bitmap(8, 8);
				tileBytes    = BitConverter.ToInt16(tileset[0], 2*i);
				tileVFlip    = (tileBytes & 0x8000) != 0;
				tileHFlip    = (tileBytes & 0x4000) != 0;
				// tilePriority = (tileBytes & 0x2000) != 0;
				tilePalette  = (tileBytes & 0x1C00) >> 10;
				tileNumber   = (tileBytes & 0x01FF);
				/*
				According to Anomie's register document, the 0x1FF to retrieve 
				"tileNumber" should be 0x3FF. When I try that, the resulting 
				maps are wrong. 0x1FF works, though. Not sure why.
				*/

				for (int row = 0; row < 8; row++)
				{
					gIndex = 0x4000 + (0x20 * tileNumber) + (2*row);
					pIndex = 0x10 * tilePalette;

					for (int col = 0; col < 8; col++)
					{
						currentPixel = 0;
						currentPixel += (graphic[gIndex + 0x00] & powersOfTwo[col]) != 0 ? 1 : 0;
						currentPixel += (graphic[gIndex + 0x01] & powersOfTwo[col]) != 0 ? 2 : 0;
						currentPixel += (graphic[gIndex + 0x10] & powersOfTwo[col]) != 0 ? 4 : 0;
						currentPixel += (graphic[gIndex + 0x11] & powersOfTwo[col]) != 0 ? 8 : 0;

						// Color zero is always transparent.
						if (currentPixel == 0)
						{
							miniTiles[i].SetPixel(col, row, Color.Transparent);
						}
						else
						{
							miniTiles[i].SetPixel(col, row, palette[pIndex + currentPixel]);
						}
					}
				}

				if (tileVFlip)  miniTiles[i].RotateFlip(RotateFlipType.RotateNoneFlipY);
				if (tileHFlip)  miniTiles[i].RotateFlip(RotateFlipType.RotateNoneFlipX);
			}


			// MEGA-TILES
			// Loop variables
			Graphics  megaTileGraphics;

			// Construct the megatiles.
			for (int i = 0; i < 0x100; i++)
			{
				megaTiles[i] = new Bitmap(16, 16);
				megaTileGraphics = Graphics.FromImage(megaTiles[i]);
				megaTileGraphics.DrawImage(miniTiles[4*i + 0], 0, 0);
				megaTileGraphics.DrawImage(miniTiles[4*i + 1], 8, 0);
				megaTileGraphics.DrawImage(miniTiles[4*i + 2], 0, 8);
				megaTileGraphics.DrawImage(miniTiles[4*i + 3], 8, 8);
			}
		}



		// ARRANGEMENT
		if (arrange[0] != null)
		{
			Graphics  mapBitmapGraphics;
			int       currentTile;


			mapBitmapGraphics = Graphics.FromImage(mapBitmap);
			for (int i = 0; i < arrangeHeight[0]; i++)
			{
				for (int j = 0; j < arrangeWidth[0]; j++)
				{
					for (int k = 0; k < 16; k++)
					{
						for (int m = 0; m < 16; m++)
						{
							currentTile = arrange[0][0x100*arrangeWidth[0]*i + 0x100*j + 0x10*k + m];

							mapBitmapGraphics.DrawImage(
							  megaTiles[currentTile],
							  (0x100 * j) + (0x10 * m),
							  (0x100 * i) + (0x10 * k)
							);
						}
					}
				}
			}
		}


		// Display the completed map.
		pbxMap.Image = mapBitmap;

		return;
	}





	void Clear(object sender, EventArgs e)
	{
		Array.Clear(graphic,    0, 0x10000);
		Array.Clear(palette,    0, 0x100);
		Array.Clear(tileset[0], 0, 0x800);
		Array.Clear(tileset[1], 0, 0x800);
		arrange[0] = null;
		arrange[1] = null;
		Array.Clear(arrangeWidth,  0, 2);
		Array.Clear(arrangeHeight, 0, 2);
		pbxMap.Image = null;

		return;
	}





	public IOGMapViewer()
	{
		graphic       = new byte[0x10000];
		palette       = new Color[0x100];
		tileset       = new byte[2][];
		tileset[0]    = new byte[0x800];
		tileset[1]    = new byte[0x800];
		arrange       = new byte[2][];
		arrangeWidth  = new int[2];
		arrangeHeight = new int[2];
		miniTiles     = new Bitmap[0x400];
		megaTiles     = new Bitmap[0x100];


		this.Text = "Illusion of Gaia Map Viewer";
		this.FormBorderStyle = FormBorderStyle.FixedSingle;
		this.MaximizeBox = false;
		this.SetClientSizeCore(
		  10 + 512 + SystemInformation.VerticalScrollBarWidth + 10,
		  10 + 40 + 10 + 512 + SystemInformation.HorizontalScrollBarHeight + 10
		);

		btnLoadROM = new Button();
		btnLoadROM.Location = new Point(10, 10);
		btnLoadROM.Size = new Size(150, 40);
		btnLoadROM.Text = "Load ROM";
		btnLoadROM.Click += new EventHandler(LoadROM);
		this.Controls.Add(btnLoadROM);

		lblMapNumber = new Label();
		lblMapNumber.Location = new Point(282, 10);
		lblMapNumber.Size = new Size(80, 20);
		lblMapNumber.Text = "Map number";
		lblMapNumber.TextAlign = ContentAlignment.MiddleCenter;
		this.Controls.Add(lblMapNumber);

		txtMapNumber = new TextBox();
		txtMapNumber.Location = new Point(282, 30);
		txtMapNumber.Size = new Size(80, 20);
		this.Controls.Add(txtMapNumber);

		btnViewMap = new Button();
		btnViewMap.Location = new Point(372, 10);
		btnViewMap.Size = new Size(70, 40);
		btnViewMap.Text = "View Map";
		btnViewMap.Click += new EventHandler(ViewMap);
		this.Controls.Add(btnViewMap);

		btnClear = new Button();
		btnClear.Location = new Point(452, 10);
		btnClear.Size = new Size(70, 40);
		btnClear.Text = "Clear";
		btnClear.Click += new EventHandler(Clear);
		this.Controls.Add(btnClear);

		pbxMap = new PictureBox();
		// DeepPink makes transparent tiles VERY visible.
		pbxMap.BackColor = Color.DeepPink;
		pbxMap.MinimumSize = new Size(512, 512);
		pbxMap.SizeMode = PictureBoxSizeMode.AutoSize;

		pnlMap = new Panel();
		pnlMap.Location = new Point(10, 60);
		pnlMap.Size = new Size(
		  512 + SystemInformation.VerticalScrollBarWidth, 
		  512 + SystemInformation.HorizontalScrollBarHeight
		);
		pnlMap.AutoScroll = true;

		pnlMap.Controls.Add(pbxMap);
		this.Controls.Add(pnlMap);
	}





	public static void Main ()
	{
		Application.Run(new IOGMapViewer());
	}
}
